from typing import Union
import warnings

import networkx as nx

from netin.utils import constants as const


def validate_int(value: int, minimum: int, maximum: int = None):
    if minimum in const.EMPTY and maximum in const.EMPTY:
        raise ValueError('At least one of minimum or maximum must be specified')
    if type(value) is not int:
        raise TypeError('value must be a int')
    if value < minimum or (maximum is not None and value > maximum):
        raise ValueError(f'Value is out of range.')


def validate_float(value: float, minimum: float, maximum: Union[None, float] = None, allow_none: bool = False):
    if value in const.EMPTY and allow_none:
        return True
    if value in const.EMPTY and not allow_none:
        raise ValueError('Value cannot be None')
    if type(value) is not float:
        raise TypeError('value must be a float')
    if minimum in const.EMPTY and maximum in const.EMPTY:
        raise ValueError('At least one of minimum or maximum must be specified')
    if value < minimum or (maximum is not None and value > maximum):
        raise ValueError(f'Value is out of range.')


def calibrate_null_probabilities(p: float) -> float:
    return const.EPSILON if p == 0 else 1 - const.EPSILON if p == 1 else p


def validate_graph_metadata(g: Union[nx.Graph, nx.DiGraph]):
    err = []
    for gkey in ['class_attribute', 'class_values', 'class_labels']:
        if gkey not in g.graph:
            err.append(gkey)
    if len(err) > 0:
        raise ValueError(f'Graph must have these attributes: "{", ".join(err)}".')

    nkey = g.graph['class_attribute']
    for n, obj in g.nodes(data=True):
        if nkey not in obj:
            raise ValueError(f'Nodes must have a "{nkey}" attribute')
        break


def validate_more_than_one(iterable):
    if len(iterable) < 2:
        raise ValueError('At least two elements are required')


def ignore_params(params: list, **kwargs) -> dict:
    tmp = set()
    for param in params:
        p = kwargs.pop(param, None)
        if p is not None:
            tmp.add(param)
    if len(tmp) > 0:
        warnings.warn(f"These parameters are ignored: {', '.join(tmp)}")

    return kwargs



from collections import defaultdict
from typing import Union, Set

import numpy as np

from netin.utils import constants as const
from netin.utils import validator as val
from .graph import Graph


class TriadicClosure(Graph):
    """Class to model triadic closure as a mechanism of edge formation given a source and a target node.

    Parameters
    ----------
    n: int
        number of nodes (minimum=2)

    f_m: float
        fraction of minorities (minimum=1/n, maximum=(n-1)/n)

    tc: float
        triadic closure probability (minimum=0, maximum=1)

    seed: object
        seed for random number generator

    Notes
    -----
    This class does not generate a graph.
    """

    ############################################################
    # Constructor
    ############################################################

    def __init__(self, n: int, f_m: float, tc: float, seed: object = None):
        Graph.__init__(self, n=n, f_m=f_m, seed=seed)
        self.tc = tc

    ############################################################
    # Init
    ############################################################

    def _infer_model_name(self):
        """
        Infers the name of the model.
        """
        return self.set_model_name(const.TC_MODEL_NAME)

    def _validate_parameters(self):
        """
        Validates the parameters of the undirected.
        """
        Graph._validate_parameters(self)
        val.validate_float(self.tc, minimum=0., maximum=1.)

    def get_metadata_as_dict(self) -> dict:
        """
        Returns the metadata (parameters) of the model as a dictionary.

        Returns
        -------
        dict
            metadata of the model
        """
        obj = Graph.get_metadata_as_dict(self)
        obj.update({
            'tc': self.tc,
        })
        return obj

    ############################################################
    # Getters & Setters
    ############################################################

    def set_triadic_closure(self, tc):
        """
        Sets the triadic closure probability `tc`.

        Parameters
        ----------
        tc: float
            triadic closure probability (minimum=0, maximum=1)
        """
        self.tc = tc

    def get_triadic_closure(self) -> float:
        """
        Returns the triadic closure probability `tc`.

        Returns
        -------
        tc: float
            triadic closure probability (minimum=0, maximum=1)
        """
        return self.tc

    ############################################################
    # Generation
    ############################################################

    def _initialize(self, class_attribute: str = 'm', class_values: list = None, class_labels: list = None):
        """
        Initializes the model.

        Parameters
        ----------
        class_attribute: str
            name of the attribute that represents the class

        class_values: list
            values of the class attribute

        class_labels: list
            labels of the class attribute mapping the class_values.
        """
        Graph._initialize(self, class_attribute, class_values, class_labels)

    def get_special_targets(self, source: int) -> object:
        """
        Returns an empty dictionary (source node ids)

        Parameters
        ----------
        source : int
            Newly added node

        Returns
        -------
        object
            Return an empty dictionary (source node ids)
        """
        return defaultdict(int)

    def get_target_probabilities_regular(self, source: Union[None, int],
                                         target_set: Union[None, Set[int]],
                                         special_targets: Union[None, object, iter] = None) -> \
            tuple[np.array, set[int]]:
        # return Graph.get_target_probabilities(self, source, target_set, special_targets)
        pass

    def get_target_probabilities(self, source: Union[None, int], target_set: Union[None, Set[int]],
                                 special_targets: Union[None, object, iter] = None) -> tuple[np.array, set[int]]:
        """
        Returns the probabilities of selecting a target node from a set of nodes based on triadic closure,
        or a regular mechanism.

        Parameters
        ----------
        source: int
            source node

        target_set: set[int]
            set of target nodes

        special_targets: object
            special targets

        Returns
        -------
        tuple[np.array, set[int]]
            probabilities of selecting a target node from a set of nodes, and the set of target nodes`
        """
        tc_prob = np.random.random()

        if tc_prob < self.tc and len(special_targets) > 0:
            target_set, probs = zip(*[(t, w) for t, w in special_targets.items()])
            probs = np.array(probs).astype(np.float32)
            probs /= probs.sum()
            target_set = set(target_set)
            return probs, target_set

        # Pick a target
        return self.get_target_probabilities_regular(source, target_set, special_targets)

    def get_target(self, source: Union[None, int],
                   targets: Union[None, Set[int]],
                   special_targets: Union[None, object, iter]) -> int:
        """
        Picks a random target node based on the homophily/preferential attachment dynamic.

        Parameters
        ----------
        source: int
            Newly added node

        targets: Set[int]
            Potential target nodes in the graph

        special_targets: object
            Special target nodes

        Returns
        -------
            int
                Target node that an edge should be created from `source`
        """
        # Collect probabilities to connect to each node in target_list
        target_set = self.get_potential_nodes_to_connect(source, targets)
        probs = self.get_target_probabilities(source, target_set, special_targets)
        return np.random.choice(a=target_set, size=1, replace=False, p=probs)[0]

    def update_special_targets(self, idx_target: int, source: int, target: int, targets: Set[int],
                               special_targets: object) -> object:
        """
        Updates the set of special targets based on the triadic closure mechanism.
        When an edge is created, multiple potential triadic closures emerge (i.e., two-hop neighbors that are not yet
        directly connected). These are added to the set of special targets.

        Parameters
        ----------
        idx_target: int
            index of the target node

        source: int
            source node

        target: int
            target node

        targets: Set[int]
            set of target nodes

        special_targets: object
            special targets

        Returns
        -------
        object
            updated special targets
        """
        if idx_target < self.k - 1:
            # Remove target candidates of source
            targets.discard(target)
            if target in special_targets:
                del special_targets[target]  # Remove target from TC candidates

            # Incr. occurrence counter for friends of new friend
            for neighbor in self.neighbors(target):
                # G[source] gives direct access (O(1)) to source's neighbors
                # G.neighbors(source) returns an iterator which would
                # need to be searched iteratively
                if neighbor not in self[source]:
                    special_targets[neighbor] += 1
        return special_targets

    ############################################################
    # Calculations
    ############################################################

    def info_params(self):
        """
        Shows the parameters of the model.
        """
        print('tc: {}'.format(self.tc))

    def info_computed(self):
        """
        Shows the computed properties of the graph.
        """
        inferred_tc = self.infer_triadic_closure()
        print("- Empirical triadic closure: {}".format(inferred_tc))

    def infer_triadic_closure(self) -> float:
        # @TODO: To be implemented
        tc = None
        return tc



############################################
# System dependencies
############################################
import operator
from typing import Tuple
from typing import Union

import netin
from . import constants as const
from .degree_rank import DegreeRank


############################################
# Class
############################################
class DegreeGroupRank(DegreeRank):
    """Sampling by degree group rank (by degree rank per class)

    Parameters
    ----------
    g: netin.Graph | netin.DiGraph
        global network

    pseeds: float
        fraction of seeds to sample

    max_tries: int
        maximum number of tries to sample a subgraph with enough classes and edges

    random_seed: object
        seed for random number generator

    kwargs: dict
        additional parameters for the sampling method

        order: str
            order of nodes by degree. Options: "asc" | "desc"

    """

    ######################################################
    # Constructor
    ######################################################
    def __init__(self, g: netin.Graph, pseeds: float, random_seed: object = None, **kwargs):
        super().__init__(g=g, pseeds=pseeds, random_seed=random_seed, **kwargs)
        self.order = self.kwargs.get("order", const.DESC)

    @property
    def method_name(self) -> str:
        name = f"{const.DEGREE_GROUP_RANK} ({const.DESC if self.is_descending() else const.ASC})"
        return name

    def sampling(self):
        super().sampling()

    def _sample(self) -> Tuple[list, Union[list, None]]:
        """
        Creates a subgraph from G based on degree rank
        """
        nodes = []
        edges = None

        ### 1. pick nodes
        _nodes = {}
        for class_value in self.g.get_class_values():
            valid = [(n, d) for n, d in self.g.degree() if d > 0 and
                     self.g.nodes[n][self.g.get_class_attribute()] == class_value]
            _nodes[class_value] = sorted(valid,
                                         key=operator.itemgetter(1),
                                         reverse=self.is_descending())
            _nodes[class_value], _ = zip(*_nodes[class_value])

        while len(nodes) < self.nseeds:
            for class_value in self.g.get_class_values():
                _nodes[class_value] = list(_nodes[class_value])
                if len(_nodes[class_value]) > 0:
                    nodes.append(_nodes[class_value].pop(0))

        num_classes = self._count_classes(nodes)

        if num_classes < const.MIN_CLASSES:
            raise ValueError(f"{num_classes} class(es). Not enough classes in the sample of {self.nseeds} nodes."
                             "Try increasing the number of seeds or sampling by DegreeGroupRank.")

        return nodes, edges



from netin import DPAH


def run():
    n = 1000
    d = 0.005
    f_m = 0.1
    plo_M = 2.0
    plo_m = 2.0
    h_MM = 0.5
    h_mm = 0.9
    seed = 1234
    g = DPAH(n=n, d=d, f_m=f_m, plo_M=plo_M, plo_m=plo_m, h_MM=h_MM, h_mm=h_mm, seed=seed)
    g.generate()
    g.info()


if __name__ == '__main__':
    run()



from .graph import Graph
from .directed import DiGraph
from .undirected import UnDiGraph
from .dh import DH
from .dpa import DPA
from .dpah import DPAH
from .pa import PA
from .pah import PAH
from .patc import PATC
from .patch import PATCH
from .tc import TriadicClosure
from .h import Homophily



import networkx as nx
import pytest

from netin import DPAH
from netin.utils import constants as const


class TestDPAH(object):

    def test_dpah_case_1(self):
        n = 200
        d = 0.1
        f_m = 0.1
        plo_M = 2.0
        plo_m = 2.0
        h_MM = 0.1
        h_mm = 0.1
        seed = 1234
        g = DPAH(n=n, d=d, f_m=f_m, plo_M=plo_M, plo_m=plo_m, h_MM=h_MM, h_mm=h_mm, seed=seed)
        g.generate()
        c1 = g.is_directed()
        c2 = g.number_of_nodes() == n
        c3 = nx.density(g) == d
        c4 = g.calculate_fraction_of_minority() == f_m
        c5 = g.get_model_name() == const.DPAH_MODEL_NAME
        assert c1 and c2 and c3 and c4 and c5, "Incorrect directed parameters."

    def test_dpah_case_2(self):
        n = 200
        d = 0.1
        f_m = 0.1
        plo_M = 2.0
        plo_m = 2.0
        h_MM = 0.9
        h_mm = 0.1
        seed = 1234
        g = DPAH(n=n, d=d, f_m=f_m, plo_M=plo_M, plo_m=plo_m, h_MM=h_MM, h_mm=h_mm, seed=seed)
        g.generate()
        c1 = g.is_directed()
        c2 = g.number_of_nodes() == n
        c3 = nx.density(g) == d
        c4 = g.calculate_fraction_of_minority() == f_m
        c5 = g.get_model_name() == const.DPAH_MODEL_NAME
        assert c1 and c2 and c3 and c4 and c5, "Incorrect directed parameters."

    def test_dpah_case_3(self):
        n = 200
        d = 0.1
        f_m = 0.1
        plo_M = 2.0
        plo_m = 2.0
        h_MM = 0.1
        h_mm = 0.9
        seed = 1234
        g = DPAH(n=n, d=d, f_m=f_m, plo_M=plo_M, plo_m=plo_m, h_MM=h_MM, h_mm=h_mm, seed=seed)
        g.generate()
        c1 = g.is_directed()
        c2 = g.number_of_nodes() == n
        c3 = nx.density(g) == d
        c4 = g.calculate_fraction_of_minority() == f_m
        c5 = g.get_model_name() == const.DPAH_MODEL_NAME
        assert c1 and c2 and c3 and c4 and c5, "Incorrect directed parameters."

    def test_dpah_case_4(self):
        n = 200
        d = 0.1
        f_m = 0.1
        plo_M = 3.0
        plo_m = 1.0
        h_MM = 0.1
        h_mm = 0.1
        seed = 1234
        with pytest.raises(ValueError, match="Value is out of range."):
            g = DPAH(n=n, d=d, f_m=f_m, plo_M=plo_M, plo_m=plo_m, h_MM=h_MM, h_mm=h_mm, seed=seed)
            g.generate()

    def test_dpah_case_5(self):
        n = 200
        d = 0.1
        f_m = 0.1
        plo_M = 1.0
        plo_m = 3.0
        h_MM = 0.1
        h_mm = 0.1
        seed = 1234
        with pytest.raises(ValueError, match="Value is out of range."):
            g = DPAH(n=n, d=d, f_m=f_m, plo_M=plo_M, plo_m=plo_m, h_MM=h_MM, h_mm=h_mm, seed=seed)
            g.generate()

    def test_dpah_case_6(self):
        n = 200
        d = 0.1
        f_m = 0.1
        plo_M = 1.0
        plo_m = 3.0
        h_MM = 0.1
        h_mm = 0.1
        seed = 1234
        with pytest.raises(TypeError, match="missing 1 required positional argument: 'plo_M'"):
            _ = DPAH(n=n, d=d, f_m=f_m, plo_m=plo_m, h_MM=h_MM, h_mm=h_mm, seed=seed)

    def test_dpah_case_7(self):
        n = 200
        d = 0.1
        f_m = 0.1
        plo_M = 1.0
        plo_m = 3.0
        h_MM = 0.1
        h_mm = 0.1
        seed = 1234
        with pytest.raises(TypeError, match="missing 1 required positional argument: 'f_m'"):
            _ = DPAH(n=n, d=d, plo_M=plo_M, plo_m=plo_m, h_MM=h_MM, h_mm=h_mm, seed=seed)

    def test_dpah_case_8(self):
        n = 200
        d = 0.1
        f_m = 0.1
        plo_M = 1.0
        plo_m = 3.0
        h_MM = 0.1
        h_mm = 0.1
        seed = 1234
        with pytest.raises(TypeError, match="missing 2 required positional arguments: 'plo_M' and 'h_mm'"):
            _ = DPAH(n=n, d=d, f_m=f_m, plo_m=plo_m, h_MM=h_MM, seed=seed)

    def test_dpah_case_9(self):
        n = 200
        d = 0.1
        f_m = 0.1
        plo_M = 1.0
        plo_m = 3.0
        h_MM = 0.1
        h_mm = 0.1
        seed = 1234
        with pytest.raises(TypeError,
                           match="missing 7 required positional arguments: 'n', 'd', 'f_m', 'plo_M', 'plo_m', 'h_MM', and 'h_mm'"):
            _ = DPAH(seed=seed)



import pytest

from netin import PATCH
from netin.utils import constants as const


class TestPATCH(object):

    def test_patch_case_1(self):
        n = 200
        k = 2
        f_m = 0.1
        h_MM = 0.1
        h_mm = 0.1
        tc = 0.5
        seed = 1234
        g = PATCH(n=n, k=k, f_m=f_m, h_MM=h_MM, h_mm=h_mm, tc=tc, seed=seed)
        g.generate()
        c1 = not g.is_directed()
        c2 = g.number_of_nodes() == n
        c3 = g.calculate_minimum_degree() == k
        c4 = g.calculate_fraction_of_minority() == f_m
        c5 = g.get_model_name() == const.PATCH_MODEL_NAME
        assert c1 and c2 and c3 and c4 and c5, "Incorrect undirected parameters."

    def test_patch_case_2(self):
        n = 200
        k = 2
        f_m = 0.1
        h_MM = 0.9
        h_mm = 0.1
        tc = 0.5
        seed = 1234
        g = PATCH(n=n, k=k, f_m=f_m, h_MM=h_MM, h_mm=h_mm, tc=tc, seed=seed)
        g.generate()
        c1 = not g.is_directed()
        c2 = g.number_of_nodes() == n
        c3 = g.calculate_minimum_degree() == k
        c4 = g.calculate_fraction_of_minority() == f_m
        c5 = g.get_model_name() == const.PATCH_MODEL_NAME
        assert c1 and c2 and c3 and c4 and c5, "Incorrect undirected parameters."

    def test_patch_case_3(self):
        n = 200
        k = 2
        f_m = 0.1
        h_MM = 0.1
        h_mm = 0.9
        tc = 0.5
        seed = 1234
        g = PATCH(n=n, k=k, f_m=f_m, h_MM=h_MM, h_mm=h_mm, tc=tc, seed=seed)
        g.generate()
        c1 = not g.is_directed()
        c2 = g.number_of_nodes() == n
        c3 = g.calculate_minimum_degree() == k
        c4 = g.calculate_fraction_of_minority() == f_m
        c5 = g.get_model_name() == const.PATCH_MODEL_NAME
        assert c1 and c2 and c3 and c4 and c5, "Incorrect undirected parameters."

    def test_patch_case_4(self):
        n = 200
        k = 2
        f_m = 0.1
        h_MM = 0.1
        h_mm = 0.1
        tc = 1.0
        seed = 1234
        g = PATCH(n=n, k=k, f_m=f_m, h_MM=h_MM, h_mm=h_mm, tc=tc, seed=seed)
        g.generate()
        c1 = not g.is_directed()
        c2 = g.number_of_nodes() == n
        c3 = g.calculate_minimum_degree() == k
        c4 = g.calculate_fraction_of_minority() == f_m
        c5 = g.get_model_name() == const.PATCH_MODEL_NAME
        assert c1 and c2 and c3 and c4 and c5, "Incorrect undirected parameters."

    def test_patch_case_5(self):
        n = 200
        k = 2
        f_m = 0.1
        h_MM = 0.1
        h_mm = 0.1
        seed = 1234
        with pytest.raises(TypeError, match="missing 1 required positional argument: 'tc'"):
            _ = PATCH(n=n, k=k, f_m=f_m, h_MM=h_MM, h_mm=h_mm, seed=seed)

    def test_patch_case_6(self):
        n = 200
        k = 2
        f_m = 0.1
        tc = 0.2
        seed = 1234
        with pytest.raises(TypeError, match="missing 2 required positional arguments: 'h_mm' and 'h_MM'"):
            _ = PATCH(n=n, k=k, f_m=f_m, tc=tc, seed=seed)

    def test_patch_case_7(self):
        n = 200
        k = 2
        f_m = 0.1
        seed = 1234
        with pytest.raises(TypeError, match="missing 3 required positional arguments: 'h_mm', 'h_MM', and 'tc'"):
            _ = PATCH(n=n, k=k, f_m=f_m, seed=seed)



from typing import Union, Set, Tuple

import numpy as np

from netin.generators.directed import DiGraph
from netin.generators.h import Homophily
from netin.utils import constants as const


class DH(DiGraph, Homophily):
    """ Creates a new DH instance. A directed graph with homophily.

    Parameters
    ----------
    n: int
        number of nodes (minimum=2)

    d: float
        edge density (minimum=0, maximum=1)

    f_m: float
        fraction of minorities (minimum=1/n, maximum=(n-1)/n)

    plo_M: float
        activity (out-degree power law exponent) majority group (minimum=1)

    plo_m: float
        activity (out-degree power law exponent) minority group (minimum=1)

    h_MM: float
        homophily within majority group (minimum=0, maximum=1)

    h_mm: float
        homophily within minority group (minimum=0, maximum=1)

    seed: object
        seed for random number generator

    Notes
    -----
    The initialization is a directed with n nodes and no edges.
    Source nodes are selected based on their activity given by plo_M (if majority) or plo_m (if minority).
    Target nodes are selected via homophily, see [Espin-Noboa2022]_.
    """

    ############################################################
    # Constructor
    ############################################################

    def __init__(self, n: int, d: float, f_m: float, plo_M: float, plo_m: float, h_MM: float, h_mm: float,
                 seed: object = None):
        DiGraph.__init__(self, n=n, d=d, f_m=f_m, plo_M=plo_M, plo_m=plo_m, seed=seed)
        Homophily.__init__(self, n=n, f_m=f_m, h_MM=h_MM, h_mm=h_mm, seed=seed)

    ############################################################
    # Init
    ############################################################

    def _infer_model_name(self):
        """
        Infers the name of the model.
        """
        return self.set_model_name(const.DH_MODEL_NAME)

    ############################################################
    # Generation
    ############################################################

    def _initialize(self, class_attribute: str = 'm', class_values: list = None, class_labels: list = None):
        """
        Initializes the model.

        Parameters
        ----------
        class_attribute: str
            name of the attribute that represents the class

        class_values: list
            values of the class attribute

        class_labels: list
            labels of the class attribute mapping the class_values.
        """
        DiGraph._initialize(self, class_attribute, class_values, class_labels)
        Homophily._initialize(self, class_attribute, class_values, class_labels)

    def get_target_probabilities(self, source: Union[None, int], target_set: Union[None, Set[int], np.array],
                                 special_targets: Union[None, object, iter] = None) -> np.array:
        """
        Returns the probabilities of the target nodes to be selected given a source node.

        Parameters
        ----------
        source: int
            source node (id)

        target_set: set
            set of target nodes (ids)

        special_targets: object
            special targets

        Returns
        -------
        probs: np.array
            probabilities of the target nodes to be selected

        """
        probs, ts = Homophily.get_target_probabilities(self, source, target_set, special_targets)
        return probs

    ############################################################
    # Calculations
    ############################################################

    def info_params(self):
        """
        Shows the parameters of the model.
        """
        DiGraph.info_params(self)
        Homophily.info_params(self)

    def info_computed(self):
        """
        Shows the computed properties of the graph.
        """
        Homophily.info_computed(self)

    def infer_homophily_values(self) -> Tuple[float, float]:
        """
        Infers analytically the homophily values for the majority and minority classes.

        Returns
        -------
        h_MM: float
            homophily within majority group

        h_mm: float
            homophily within minority group
        """
        from sympy import symbols
        from sympy import Eq
        from sympy import solve

        f_m = self.calculate_fraction_of_minority()
        f_M = 1 - f_m

        e = self.calculate_edge_type_counts()
        e_MM = e['MM']
        e_mm = e['mm']
        e_Mm = e['Mm']
        e_mM = e['mM']

        p_MM = e_MM / (e_MM + e_Mm)
        p_mm = e_mm / (e_mm + e_mM)

        # equations
        hmm, hMM, hmM, hMm = symbols('hmm hMM hmM hMm')
        eq1 = Eq((f_m * hmm) / ((f_m * hmm) + (f_M * hmM)), p_mm)
        eq2 = Eq(hmm + hmM, 1)

        eq3 = Eq((f_M * hMM) / ((f_M * hMM) + (f_m * hMm)), p_MM)
        eq4 = Eq(hMM + hMm, 1)

        solution = solve((eq1, eq2, eq3, eq4), (hmm, hmM, hMM, hMm))
        h_MM, h_mm = solution[hMM], solution[hmm]
        return h_MM, h_mm

    def _makecopy(self):
        """
        Makes a copy of the current object.
        """
        return self.__class__(n=self.n,
                              d=self.d,
                              f_m=self.f_m,
                              plo_M=self.plo_M,
                              plo_m=self.plo_m,
                              h_MM=self.h_MM,
                              h_mm=self.h_mm,
                              seed=self.seed)



from typing import Union, Set, List, Tuple

import numpy as np
import pandas as pd
import powerlaw


def get_pdf(df: pd.DataFrame, x: str, total: float) -> Tuple[np.ndarray, np.ndarray]:
    """Computes the probability density of the input data.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame that contains the data.
    x : str
        The column name of the data.
    total : float
        The total amount by which to normalize the data.

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        Two arrays holding the x values and y values (their probability).
    """
    values = df.groupby(x).size()
    xs = values.index.values
    ys = values.values / total
    return xs, ys


def get_cdf(df: pd.DataFrame, x: str, total: float = None) -> (np.ndarray, np.ndarray):
    """Computes the cumulative distribution CDF of the input data.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame that contains the data.
    x : str
        The column name of the data.
    total : float
        The total amount by which to normalize the data. (not used here)

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        Two arrays holding the x values and the y values (CDF)
    """
    xs = np.sort(df[x].values)
    n = xs.size
    ys = np.arange(1, n + 1) / n
    return xs, ys


def get_ccdf(df: pd.DataFrame, x: str, total: float = None) -> (np.ndarray, np.ndarray):
    """Computes the complementary cumulative distribution CCDF of the input data.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame that contains the data.
    x : str
        The column name of the data.
    total : float
        The total amount by which to normalize the data.

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        Two arrays holding the x values and the y values (CCDF)
    """
    xs, ys = get_cdf(df, x, total)
    return xs, 1 - ys


def get_disparity(df: pd.DataFrame, x: str, total: float = None) -> (np.ndarray, np.ndarray):
    """Computes the disparity of the input data given by the column `x`.

    Parameters
    ----------
    df: pd.DataFrame
        DataFrame that contains the data.

    x: str
        The column name of the data.

    total: float
        The total amount by which to normalize the data. (not used here)

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        Two arrays holding the x values (ranking) and the y values (disparity)
    """
    from netin.stats import ranking
    from netin.utils import constants as const

    gx, gy = get_gini_coefficient(df, x, total)
    fx, fy = get_fraction_of_minority(df, x, total)
    f_m = df.query("class_label == @const.MINORITY_LABEL").shape[0] / df.shape[0]

    inequality_y = ranking.get_ranking_inequality(gy)
    inequity_x = ranking.get_ranking_inequity(f_m, fy)
    return inequity_x, inequality_y


def get_fraction_of_minority(df: pd.DataFrame, x: str, total: float = None) -> (np.ndarray, np.ndarray):
    """Computes the fraction of minority in each top-k rank.

    Parameters
    ----------
    df: pd.DataFrame
        DataFrame that contains the data.

    x: str
        The column name of the data.

    total: float
        The total amount by which to normalize the data. (not used here)

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        Two arrays holding the x values (ranking) and the y values (fraction of minority)
    """
    from netin.stats import ranking
    xs, ys = ranking.get_fraction_of_minority_in_ranking(df, x)
    return xs, ys


def get_gini_coefficient(df: pd.DataFrame, x: str, total: float = None) -> (np.ndarray, np.ndarray):
    """Computes the Gini coefficient of the distribution in each top-k rank.

    Parameters
    ----------
    df: pd.DataFrame
        DataFrame that contains the data.

    x: str
        The column name of the data.

    total: float
        The total amount by which to normalize the data. (not used here)

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        Two arrays holding the x values (ranking) and the y values (Gini coefficient)
    """
    from netin.stats import ranking
    xs, ys = ranking.get_gini_in_ranking(df, x)
    return xs, ys


def fit_power_law(data: Union[np.array, Set, List], discrete: bool = True,
                  xmin: Union[None, int, float] = None, xmax: Union[None, int, float] = None, **kwargs) -> powerlaw.Fit:
    """Fits a power-law of a given distribution.

    Parameters
    ----------
    data: Union[np.array, Set, List]
        The data to fit.

    discrete: bool
        Whether the data is discrete or not.

    xmin: Union[None, int, float]
        The minimum value of the data.

    xmax: Union[None, int, float]
        The maximum value of the data.

    kwargs: dict
        Additional arguments to pass to the powerlaw.Fit constructor.

    Returns
    -------
    powerlaw.Fit
        The fitted power-law.
    """
    fit = powerlaw.Fit(data, discrete=discrete, xmax=xmax, xmin=xmin, **kwargs)
    return fit



"""
NetIn is a Python package for the analysis of network inequalities.
It is based on the NetworkX package and provides a set of functions to study inequalities (e.g., in ranking, inference)
in social networks.
"""

__version__ = '1.0.5.8.8'

from netin import utils
from netin.utils import *

from netin import generators
from netin.generators import *

from netin import algorithms
from netin.algorithms import sampling



from netin import DH
from netin import DPA
from netin import DPAH
from netin.utils import constants as const


class TestDiGraph(object):

    def test_case_dpa(self):
        n = 200
        d = 0.01
        f_m = 0.1
        plo_M = 2.0
        plo_m = 2.0
        seed = 5678
        g = DPA(n=n, d=d, f_m=f_m, plo_M=plo_M, plo_m=plo_m, seed=seed)
        g.generate()
        c1 = g.is_directed()
        c2 = g.number_of_nodes() == n
        c3 = g.number_of_edges() == g.get_expected_number_of_edges()
        c4 = g.calculate_fraction_of_minority() == f_m
        c5 = g.get_model_name() == const.DPA_MODEL_NAME
        assert c1 and c2 and c3 and c4 and c5, "Incorrect directed parameters."

    def test_case_dh(self):
        n = 200
        d = 0.01
        f_m = 0.1
        h_MM = 0.9
        h_mm = 0.9
        plo_M = 2.0
        plo_m = 2.0
        seed = 5678
        g = DH(n=n, d=d, f_m=f_m, h_MM=h_MM, h_mm=h_mm, plo_M=plo_M, plo_m=plo_m, seed=seed)
        g.generate()
        c1 = g.is_directed()
        c2 = g.number_of_nodes() == n
        c3 = g.number_of_edges() == g.get_expected_number_of_edges()
        c4 = g.calculate_fraction_of_minority() == f_m
        c5 = g.get_model_name() == const.DH_MODEL_NAME
        assert c1 and c2 and c3 and c4 and c5, "Incorrect directed parameters."

    def test_case_dpah(self):
        n = 200
        d = 0.01
        f_m = 0.1
        h_MM = 0.9
        h_mm = 0.9
        plo_M = 2.0
        plo_m = 2.0
        seed = 5678
        g = DPAH(n=n, d=d, f_m=f_m, h_MM=h_MM, h_mm=h_mm, plo_M=plo_M, plo_m=plo_m, seed=seed)
        g.generate()
        c1 = g.is_directed()
        c2 = g.number_of_nodes() == n
        c3 = g.number_of_edges() == g.get_expected_number_of_edges()
        c4 = g.calculate_fraction_of_minority() == f_m
        c5 = g.get_model_name() == const.DPAH_MODEL_NAME
        assert c1 and c2 and c3 and c4 and c5, "Incorrect directed parameters."

    def test_case_all(self):
        n = 200
        d = 0.01
        f_m = 0.1
        h_MM = 0.9
        h_mm = 0.9
        plo_M = 2.0
        plo_m = 2.0
        seed = 1234
        g_dpa = DPA(n=n, d=d, f_m=f_m, plo_M=plo_M, plo_m=plo_m, seed=seed)
        g_dh = DH(n=n, d=d, f_m=f_m, h_MM=h_MM, h_mm=h_mm, plo_M=plo_M, plo_m=plo_m, seed=seed)
        g_dpah = DPAH(n=n, d=d, f_m=f_m, h_MM=h_MM, h_mm=h_mm, plo_M=plo_M, plo_m=plo_m, seed=seed)

        g_dpa.generate()
        g_dh.generate()
        g_dpah.generate()

        c1 = g_dpa.number_of_nodes() == g_dh.number_of_nodes() == g_dpah.number_of_nodes() == n
        c2 = g_dpa.calculate_fraction_of_minority() == g_dh.calculate_fraction_of_minority() == g_dpah.calculate_fraction_of_minority() == f_m
        c3 = g_dpa.get_expected_number_of_edges() == g_dh.get_expected_number_of_edges() == g_dpah.get_expected_number_of_edges()

        c4 = g_dpa.get_model_name() == const.DPA_MODEL_NAME
        c5 = g_dh.get_model_name() == const.DH_MODEL_NAME
        c6 = g_dpah.get_model_name() == const.DPAH_MODEL_NAME
        assert c1 and c2 and c3 and c4 and c5 and c6, "Incorrect directed parameters."



from typing import Union, Set, Tuple

import numpy as np
import networkx as nx

from netin.utils import constants as const
from netin.generators.tc import TriadicClosure
from .pa import PA


class PATC(PA, TriadicClosure):
    """Creates a new PATC instance. An undirected graph with preferential attachment and triadic closure.

    Parameters
    ----------
    n: int
        number of nodes (minimum=2)

    k: int
        minimum degree of nodes (minimum=1)

    f_m: float
        fraction of minorities (minimum=1/n, maximum=(n-1)/n)

    tc: float
        probability of a new edge to close a triad (minimum=0, maximum=1.)

    seed: object
        seed for random number generator

    Notes
    -----
    The initialization is an undirected with n nodes and no edges.
    Then, everytime a node is selected as source, it gets connected to k target nodes.
    Target nodes are selected via preferential attachment `in-degree` [BarabasiAlbert1999]_, or
    triadic closure `tc` [HolmeKim2002]_.
    """

    ############################################################
    # Constructor
    ############################################################

    def __init__(self, n: int, k: int, f_m: float, tc: float, seed: object = None):
        PA.__init__(self, n=n, k=k, f_m=f_m, seed=seed)
        TriadicClosure.__init__(self, n=n, f_m=f_m, tc=tc, seed=seed)

    ############################################################
    # Init
    ############################################################

    def _infer_model_name(self):
        """
        Infers the name of the model.
        """
        return self.set_model_name(const.PATC_MODEL_NAME)

    def _validate_parameters(self):
        """
        Validates the parameters of the undirected.
        """
        PA._validate_parameters(self)
        TriadicClosure._validate_parameters(self)

    def get_metadata_as_dict(self) -> dict:
        """
        Returns the metadata information (input parameters of the model) of the graph as a dictionary.

        Returns
        -------
        dict
            Dictionary with the graph's metadata
        """
        obj1 = PA.get_metadata_as_dict(self)
        obj2 = TriadicClosure.get_metadata_as_dict(self)
        obj1.update(obj2)
        return obj1

    ############################################################
    # Generation
    ############################################################

    def info_params(self):
        """
        Shows the parameters of the model.
        """
        PA.info_params(self)
        TriadicClosure.info_params(self)

    def get_special_targets(self, source: int) -> object:
        """
        Returns the initial set of special targets based on the triadic closure mechanism.

        Parameters
        ----------
        source : int
             Newly added node

        Returns
        -------
        object
            Empty dictionary (source node ids)

        See Also
        --------
        :py:meth:`get_special_targets() <TriadicClosure.get_special_targets>` in :class:`netin.TC`.
        """
        return TriadicClosure.get_special_targets(self, source)

    def get_target_probabilities(self, source: Union[None, int], target_set: Union[None, Set[int]],
                                 special_targets: Union[None, object, iter] = None) -> tuple[np.array, set[int]]:
        """
        Returns the probabilities of selecting a target node from a set of nodes based on the preferential attachment
        or triadic closure.

        Parameters
        ----------
        source: int
            source node

        target_set: set[int]
            set of target nodes

        special_targets: object
            special targets

        Returns
        -------
        tuple[np.array, set[int]]
            probabilities of selecting a target node from a set of nodes, and the set of target nodes

        See Also
        --------
        :py:meth:`get_target_probabilities() <TriadicClosure.get_target_probabilities>` in :class:`netin.TC`.
        """
        return TriadicClosure.get_target_probabilities(self, source, target_set, special_targets)

    def get_target_probabilities_regular(self, source: Union[None, int],
                                         target_set: Union[None, Set[int]],
                                         special_targets: Union[None, object, iter] = None) -> \
            Tuple[np.array, Set[int]]:
        """
        Returns the probabilities of selecting a target node from a set of nodes based on the preferential attachment.

        Parameters
        ----------
        source: int
            source node

        target_set: set[int]
            set of target nodes

        special_targets: object
            special targets

        Returns
        -------
        tuple[np.array, set[int]]
            probabilities of selecting a target node from a set of nodes, and the set of target nodes

        See Also
        --------
        :py:meth:`get_target_probabilities() <PA.get_target_probabilities>` in :class:`netin.PA`.
        """
        return PA.get_target_probabilities(self, source, target_set, special_targets)

    def update_special_targets(self, idx_target: int, source: int, target: int, targets: Set[int],
                               special_targets: object) -> object:
        """
        Updates the set of special targets based on the triadic closure mechanism.

        Parameters
        ----------
        idx_target: int
            index of the target node

        source: int
            source node

        target: int
            target node

        targets: Set[int]
            set of target nodes

        special_targets: object
            special targets

        Returns
        -------
        object
            updated special targets

        See Also
        --------
        :func:`netin.TC.update_special_targets`

        """
        return TriadicClosure.update_special_targets(self, idx_target, source, target, targets, special_targets)

    ############################################################
    # Calculations
    ############################################################

    def infer_triadic_closure(self) -> float:
        """
        Approximates analytically the triadic closure value of the graph.

        Returns
        -------
        float
            triadic closure probability of the graph
        """
        tc = nx.average_clustering(self)
        return tc

    def _makecopy(self):
        """
        Makes a copy of the current object.
        """
        return self.__class__(n=self.n,
                              k=self.k,
                              f_m=self.f_m,
                              tc=self.tc,
                              seed=self.seed)



# Copyright (c) 2023 NetInCSH.  All rights reserved.
# This work is licensed under a Creative Commons Attribution-NonCommercial-ShareAlike 4.0 International
# License.  (https://creativecommons.org/licenses/by-nc-sa/4.0/legalcode

import os
import sys
from glob import glob

from setuptools import setup

with open("netin/__init__.py") as fid:
    for line in fid:
        if line.startswith("__version__"):
            version = line.strip().split()[-1][1:-1]
            break

if sys.version_info[:2] < (3, 8):
    error = (f"NetIn {version} requires Python 3.9 or later ({sys.version_info[:2]} detected). \n")
    sys.stderr.write(error + "\n")
    sys.exit(1)

name = "netin"
description = "Python package to study inequalities in social networks"
authors = {
    "Karimi": ("Fariba Karimi", "karimi@csh.ac.at"),
    "Espín-Noboa": ("Lisette Espín-Noboa", "espin@csh.ac.at"),
    "Bachmann": ("Jan Bachmann", "bachmann@csh.ac.at"),
}
maintainer = "NetIn Developers"
maintainer_email = "netin-dev@googlegroups.com"
url = "https://github.com/CSHVienna/NetworkInequalities"
platforms = ["Linux", "Mac OSX", "Windows", "Unix"]
keywords = [
    "Networks",
    "Inequalities",
    "Social Networks",
    "Ranking",
    "Inference"
    "Graph Theory",
    "Mathematics",
    "network",
    "undirected",
    "discrete mathematics",
    "math",
]
classifiers = [
    "Development Status :: 3 - Alpha",
    "Intended Audience :: Developers",
    "Intended Audience :: Education",
    "Intended Audience :: Information Technology",
    "Intended Audience :: Science/Research",
    "Operating System :: OS Independent",
    "Programming Language :: Python :: 3.9",
    "Topic :: Software Development :: Libraries :: Python Modules",
    "Topic :: Scientific/Engineering",
    "Topic :: Scientific/Engineering :: Information Analysis",
    "Topic :: Scientific/Engineering :: Mathematics",
    "Topic :: Scientific/Engineering :: Physics",
]

packages = [
    "netin",
    "netin.algorithms.sampling",
    "netin.generators",
    "netin.utils",
    "netin.stats",
    "netin.viz",
    "netin.generators.tests",
]

docdirbase = "share/doc/netin-%s" % version
# add basic documentation
data = [(docdirbase, glob("*.txt"))]
# add examples
for d in [
    ".",
    "advanced",
    "algorithms",
    "basic",
    "drawing",
    "undirected",
    "subclass",
]:
    dd = os.path.join(docdirbase, "examples", d)
    pp = os.path.join("examples", d)
    data.append((dd, glob(os.path.join(pp, "*.txt"))))
    data.append((dd, glob(os.path.join(pp, "*.py"))))
    data.append((dd, glob(os.path.join(pp, "*.bz2"))))
    data.append((dd, glob(os.path.join(pp, "*.gz"))))
    data.append((dd, glob(os.path.join(pp, "*.mbox"))))
    data.append((dd, glob(os.path.join(pp, "*.edgelist"))))
# add js force examples
dd = os.path.join(docdirbase, "examples", "javascript/force")
pp = os.path.join("examples", "javascript/force")
data.append((dd, glob(os.path.join(pp, "*"))))

# add the tests subpackage(s)
package_data = {
    "netin": ["tests/*.py"],
    "netin.generators": ["tests/*.py"],
}


def parse_requirements_file(filename):
    with open(filename) as fid:
        requires = [l.strip() for l in fid.readlines() if not l.startswith("#")]
    return requires


install_requires = parse_requirements_file("requirements/default.txt")

extras_require = {
    dep: parse_requirements_file("requirements/" + dep + ".txt")
    for dep in ["default", "test"]  # , "developer", "doc", "extra"]
}

with open("README.rst") as fh:
    long_description = fh.read()

if __name__ == "__main__":
    setup(
        name=name,
        version=version,
        maintainer=maintainer,
        maintainer_email=maintainer_email,
        author=authors["Espín-Noboa"][0],
        author_email=authors["Espín-Noboa"][1],
        description=description,
        keywords=keywords,
        long_description=long_description,
        long_description_content_type="text/x-rst",
        platforms=platforms,
        url=url,
        classifiers=classifiers,
        packages=packages,
        data_files=data,
        package_data=package_data,
        install_requires=install_requires,
        extras_require=extras_require,
        python_requires=">=3.9",
        zip_safe=False,
    )



############################################
# System dependencies
############################################
import numpy as np

import netin
from . import constants as const
from .sampling import Sampling


############################################
# Class
############################################
class RandomEdges(Sampling):
    """Random edge sampling.

    Parameters
    ----------
    g: netin.Graph | netin.DiGraph
        global network

    pseeds: float
        fraction of seeds to sample

    max_tries: int
        maximum number of tries to sample a subgraph with enough classes and edges

    random_seed: object
        seed for random number generator

    kwargs: dict
        additional parameters for the sampling method
    """

    ######################################################
    # Constructor
    ######################################################
    def __init__(self, g: netin.Graph, pseeds: float, max_tries: int = const.MAX_TRIES,
                 random_seed: object = None, **kwargs):
        super().__init__(g, pseeds, max_tries, random_seed, **kwargs)

    @property
    def method_name(self) -> str:
        return const.RANDOM_EDGES

    def sampling(self):
        super().sampling()

    def _sample(self):
        """
        Creates a subgraph from G based on random edge sampling
        """
        num_classes = 0
        nodes = None
        edges = None

        ### 1. pick random edges
        while num_classes < const.MIN_CLASSES:
            tmp_edges = list(self.g.edges())
            np.random.shuffle(tmp_edges)
            edges = set()
            nodes = set()

            while len(nodes) < self.nseeds and len(tmp_edges) > 0:
                edge = tmp_edges.pop(0)
                nodes |= set(edge)
                edges.add(edge)

            num_classes = self._count_classes(nodes)

        return nodes, edges



import numpy as np

# GENERAL

EMPTY = [None, np.nan]

# GRAPHS

NO_HOMOPHILY = [0.5, None, np.nan]
NO_TRIADIC_CLOSURE = [0, None, np.nan]

CLASS_ATTRIBUTE = 'm'

MAJORITY_LABEL = 'M'
MINORITY_LABEL = 'm'
CLASS_LABELS = [MAJORITY_LABEL, MINORITY_LABEL]

MAJORITY_VALUE = 0
MINORITY_VALUE = 1
CLASS_VALUES = [MAJORITY_VALUE, MINORITY_VALUE]

# GENERATIVE MODELS

EPSILON = 0.00001
MAX_TRIES_EDGE = 100

DIRECTED_MODEL_NAME = 'Directed'
UNDIRECTED_MODEL_NAME = 'Undirected'

H_MODEL_NAME = 'Homophily'
TC_MODEL_NAME = 'Triadic Closure'

PA_MODEL_NAME = 'PA'
PAH_MODEL_NAME = 'PAH'
PATC_MODEL_NAME = 'PATC'
PATCH_MODEL_NAME = 'PATCH'

DH_MODEL_NAME = 'DH'
DPA_MODEL_NAME = 'DPA'
DPAH_MODEL_NAME = 'DPAH'

HOMOPHILY_MODELS = [PAH_MODEL_NAME, PATCH_MODEL_NAME, DH_MODEL_NAME, DPAH_MODEL_NAME, H_MODEL_NAME]

# NODE METRICS

VALID_METRICS = ['degree', 'in_degree', 'out_degree', 'clustering',
                 'betweenness', 'closeness', 'eigenvector', 'pagerank']

# RANKING

RANK_RANGE = np.arange(0.1, 1 + 0.1, 0.1).astype(np.float32)
INEQUITY_BETA = 0.05
INEQUITY_OVER = 'over-represented'
INEQUITY_UNDER = 'under-represented'
INEQUITY_FAIR = 'fair'
INEQUALITY_CUTS = [0.3, 0.6]
INEQUALITY_HIGH = 'skewed'
INEQUALITY_MODERATE = 'moderate'
INEQUALITY_LOW = 'equality'
